// DATES.CPP
//
// DATES module
//
//   Author: Cal Lenders  
//   Rev 1.0
//

#include "dates.h"

const int daysInMonth[12] = {
    31, 28, 31, 30, 31, 30,
    31, 31, 30, 31, 30, 31
};

// Days from the beginning of the year
//   to the beginning of the month:

const int daysToMonth[12] = {
     0,  31,  59,  90, 120, 151,
   181, 212, 243, 273, 304, 334
};

const char *monthName[12] = {
    "January", "February", "March",
    "April", "May", "June",
    "July", "August", "September",
    "October", "November", "December"
};

//****************************************************************

bool LeapYear (int year)

// year must be between 1900 and 2999.
// Returns true if year is a leap year.

{
    return ((year % 4 == 0 && year % 100 != 0) ||
         year % 400 == 0);
}

//****************************************************************

bool ValidDate(const DATE &date)

// Returns true if date contains a valid date between
// Jan 1, 1990 and Dec 31, 2100.

{
    int days = daysInMonth[date.month - 1];
    if (date.month == 2 && LeapYear(date.year)) days++;

    return (date.month >= 1 && date.month <= 12 &&
                date.day >= 1 && date.day <= days &&
                date.year >= 1900 && date.year <= 2100);
}

//****************************************************************

static int DaysSince1900(const DATE &date)

// date contains a valid date
//   (between Jan 1, 1990 and Dec 31, 2100).
// Returns number of days passed since Jan 1, 1990.

{
    int years, days;

    years = date.year - 1900;
    days = years * 365;

    // Correct for the previous leap years, excluding 1900 but
    // including 2000, and the current leap year.
    if (years >= 5) days += (years - 5) / 4;
    if (LeapYear(date.year) && date.month > 2) days++;

    days += daysToMonth[date.month - 1] +   date.day - 1;

    return days;
}

//****************************************************************

int DaysBetween(const DATE &date1, const DATE &date2)

// date1 and date2 must contain valid dates.
// Returns number of days passed from date1 to date2.

{
    return DaysSince1900(date2) - DaysSince1900(date1);
}

//****************************************************************

bool Earlier(const DATE &date1, const DATE &date2)

// date1 and date2 must contain valid dates.
// Returns true if date1 is earlier than date2, false otherwise.

{
    return DaysBetween(date1, date2) > 0;
}

//****************************************************************

void AddDays(const DATE &date1, DATE &date2, int days)

// date1 contains a valid date and days is >= 0.
// Fills date2 with a new date obtained by
//   adding days to date1.

{
    date2 = date1;

    while (days > 0) {
        int monthLength = daysInMonth[date2.month - 1];
        if (date2.month == 2 && LeapYear(date2.year)) monthLength++;
        if (date2.day + days > monthLength) {
            date2.month++;
            if (date2.month > 12) {
                date2.month = 1;
                date2.year++;
            }
            days -= monthLength - date2.day + 1;
            date2.day = 1;
        }
        else {
            date2.day += days;
            days = 0;
        }
    }
}

//****************************************************************

ostream &operator<< (ostream &os, const DATE &date)

// Usage:
//   cout << date;
// Prints the date, as in "January 1, 1990".

{
    os << monthName[date.month - 1] << ' '
       << date.day << ", "
       << date.year;

    return os;
} 

//****************************************************************

