// HEAP.H

// Heap template class
// Author: Q. Prior

#ifndef _HEAP_H_
#define _HEAP_H_

#include "bool.h"
#include "apvector.h"

template <class itemType>
class HEAP {

  private:

    int mSize;                    // Maximum size
    int mNodes;                   // Current number of nodes
    apvector<itemType> mBuffer;   // Array of heap elements

  public:

    HEAP (int size=1024);  // Constructor
    ~HEAP();
    bool isEmpty();
    bool isFull();
    bool Insert(const itemType &x);
    bool Remove(itemType &x);
};

//****************************************************************

template <class itemType>
HEAP<itemType>::HEAP(int size)

// Constructor: builds a heap of size "size".

        : mNodes(0), mSize(size), mBuffer(size + 1)
        // Buffer is size+1 because the first element is not used.
{}

template <class itemType>
HEAP<itemType>::~HEAP()

// Destructor.

{}

//****************************************************************

template <class itemType>
bool HEAP<itemType>::isEmpty()

{
    return mNodes == 0;
}

template <class itemType>
bool HEAP<itemType>::isFull()
{
    return mNodes == mSize;
}

//****************************************************************

template <class itemType>
bool HEAP<itemType>::Insert(const itemType &x)

// Inserts x into the heap.
// Returns true if successful, false if the heap is full.

{
    if (isFull())
        return false;

    mNodes++;  // The last node of the heap is now vacant.

    // Starting from the (vacant) last node, go from node i to
    //   its parent iParent and, as long as the parent is
    //   smaller than x, move the parent down:

    int i = mNodes;
    int iParent;

    while (i > 1) {
        iParent = i/2;
        if (x <= mBuffer[iParent])
            break;
        mBuffer[i] = mBuffer[iParent]; // Move the parent down;
        i = iParent;                   // mBuffer[i] is now vacant.
    }

    // Insert x into the created vacancy:
    mBuffer[i] = x;

    return true;
}

//****************************************************************

template <class itemType>
bool HEAP<itemType>::Remove(itemType &x)

// Removes the largest element (from the root of the heap).
// Returns true if successful, false if the heap is empty.

{
    if (isEmpty())
        return false;

    // Retrieve the top element:

    x = mBuffer[1];             

    // Starting from the vacant root, go from node iParent to its
    //   larger child i and, as long as that child
    //   is greater than the last element of the heap,
    //   move that child up:

    int iParent = 1;       // root
    int i = 2;             // its left child

    while (i <= mNodes) {
        // Set i to the right child, i+1, if it
        //   exists and is larger:
        if (i < mNodes && mBuffer[i] < mBuffer[i+1]) i++;

        // Compare with the last node:
        if (mBuffer[i] <= mBuffer[mNodes]) break;

        mBuffer[iParent] = mBuffer[i];   // Move the child up; 
        iParent = i;             // mBuffer[iParent] is now vacant.
        i *= 2;                  // i is set to its left child
    }

    // Move the last element into the created vacancy:
    if (mNodes > 1)
        mBuffer[iParent] = mBuffer[mNodes];
    mNodes--;
 
    return true;
}

#endif // _HEAP_H_

