// SORTS.CPP
//
// Implementations of the binary search and sorting algorithms.
//

#include "apvector.h"

//****************************************************************
//***********              Binary search            **************
//****************************************************************

template <class SOMETYPE>
int BinarySearch (const apvector<SOMETYPE> &v, SOMETYPE target)

// Performs binary search in the array v.
//   The elements in the array must be arranged in
//   ascending order.
//   Looks for a value equal to target.
//   (the ==, < and > operators as well as assignment must be
//   defined for SOMETYPE.)
// Returns: the location of target in v, if found; -1 otherwise.

{
    int location = -1;
    int left = 0, right = v.length() - 1, middle;

    while (left <= right) {
        middle = (left + right) / 2;
        if (target > v[middle])
            left = middle + 1;
        else if (target < v[middle])
            right = middle - 1;
        else {
            // if (target == v[middle])
            location = middle;
            break;
        }
    }
    return location;
}

//****************************************************************
//***********                 SORTING               **************
//****************************************************************

template <class SOMETYPE>
inline void Swap(SOMETYPE &x, SOMETYPE &y)

{
    SOMETYPE temp = x; x = y; y = temp;
}

//****************************************************************
//***********              Selection sort           **************
//****************************************************************

template <class SOMETYPE>
void SelectionSort (apvector<SOMETYPE> &v)

{
    int i, iMax, n = v.length();

    while (n > 1) {

        // Find the largest element:
        for (iMax = 0, i = 1;   i < n;   i++)
            if (v[i] > v[iMax]) iMax = i;

        // Swap it with the last element:
        Swap(v[iMax], v[n-1]);

        n--;
    }
}

//****************************************************************
//***********              Insertion sort           **************
//****************************************************************

template <class SOMETYPE>
void InsertionSort (apvector<SOMETYPE> &v)

{
    int i, j, n = v.length();
    SOMETYPE vCurrent;

    for (i = 1;   i < n;   i++) {

        // Save the current element:
        vCurrent = v[i];

        // Find location j where it should be inserted
        //   among the first i-1 elements:
        for (j = 0;   j < i;  j++)
            if (v[j] >= vCurrent)
                break;

        // Shift all elements between j and i one place to the right:
        for (int k = i-1;  k >= j;  k--)
            v[k+1] = v[k];

        // Insert saved element where it belongs:
        v[j] = vCurrent;
    }
}

//****************************************************************

template <class SOMETYPE>
void InsertionSort2 (apvector<SOMETYPE> &v)

// A variation of insertion sort
//   with scanning the sublist backwards from i-1 down to 0
 
{
    int i, j, n = v.length();
    SOMETYPE vCurrent;

    for (i = 1;   i < n;   i++) {

        // Save the current element:
        vCurrent = v[i];

        // Going backwards from v[i-1], shift elements to the
        //   right until you find an element v[j] <= vCurrent:
        for (j = i-1;   j >= 0;  j--) {
            if (v[j] <= vCurrent)
                break;
            v[j+1]  = v[j];
        }

        // Insert saved element after v[j]:
        v[j+1] = vCurrent;
    }
}

//****************************************************************
//***********              Bubble sort              **************
//****************************************************************

template <class SOMETYPE>
void BubbleSort (apvector<SOMETYPE> &v)

// Quits early when the array is confirmed to be in order.

{
    int i, n = v.length(), disorder = n;

    while (disorder) {

        disorder = 0;

        // Bubble up and mark any disorder:
        //   (When n becomes 1, the "for" loop is not
        //    executed, "disorder" remains 0, and the "while"
        //    loop is broken.)

        for (i = 1;   i < n;   i++) {
            if (v[i] < v[i-1]) {
                Swap(v[i], v[i-1]);
                disorder++;
            }
        }
        n--;
    }
}

//****************************************************************
//***********               Mergesort               **************
//****************************************************************

// MERGESRT.H

template <class SOMETYPE>
class MERGESORT {

  public:

    void Sort (apvector<SOMETYPE> &v);

  private:

    apvector<SOMETYPE> work;

    void RecursiveSort(apvector<SOMETYPE> &v, int i1, int i2);
    void Merge(apvector<SOMETYPE> &v, int i1, int m, int i2);
};

//****************************************************************

template <class SOMETYPE>
void MERGESORT<SOMETYPE>::Sort (apvector<SOMETYPE> &v)

// Sorts array v in ascending order.

{
    int n = v.length();
    work.resize(n);
    RecursiveSort(v, 0, n-1);
    work.resize(0);
}

template <class SOMETYPE>
void MERGESORT<SOMETYPE>::RecursiveSort (apvector<SOMETYPE> &v,
                                                 int i1, int i2)

// Sorts elements of v between i1 and i2 in ascending order.

{
    int m;

    if (i2 - i1 <= 1) {           // Base case:
        if (i2 - i1 == 1 && v[i2] < v[i1])
            Swap(v[i2], v[i1]);
    }
    else {                        // Recursive case:
        m = (i2 + i1) / 2;
        RecursiveSort(v, i1, m);
        RecursiveSort(v, m + 1, i2);
        Merge(v, i1, m, i2);
    }
}

//****************************************************************

template <class SOMETYPE>
void MERGESORT<SOMETYPE>::
        Merge(apvector<SOMETYPE> &v, int i1, int m, int i2)

// Merges sorted segments of the array v between i1 and m
//   and between m+1 and i2 into one sorted segment between i1 and i2.

{
    int j1 = i1, j2 = m + 1, j;

    // Merge two arrays into the work array:
    for (j = i1;   j <= i2;   j++) {
        if (j1 <= m && j2 <= i2) {
            if (v[j1] < v[j2])
                work[j] = v[j1++];
            else
                work[j] = v[j2++];
        }
        else if (j1 <= m)
            work[j] = v[j1++];
        else // if (j2 <= i2)
            work[j] = v[j2++];
    }

    // Copy work back to v:
    for (j = i1;   j <= i2;   j++)
        v[j] = work[j];
}

//****************************************************************
//***********               Quicksort               **************
//****************************************************************

// QUICKSRT.H

template <class SOMETYPE>
class QUICKSORT {

  public:

    void Sort (apvector<SOMETYPE> &v);

  private:

    void RecursiveSort(apvector<SOMETYPE> &v, int i1, int i2);
    int Split(apvector<SOMETYPE> &v, int i1, int i2);
        // Returns the new location of the pivot element.

};

//****************************************************************

template <class SOMETYPE>
void QUICKSORT<SOMETYPE>::Sort (apvector<SOMETYPE> &v)

// Sorts array v in ascending order.

{
    int n = v.length();
    RecursiveSort(v, 0, n-1);
}

#ifdef SIMPLIFIED

template <class SOMETYPE>
void QUICKSORT<SOMETYPE>::RecursiveSort (apvector<SOMETYPE> &v,
                                                 int i1, int i2)

// Sorts elements of v between i1 and i2 in ascending order.

{
    int p;

    if (i2 - i1 <= 1) {  // Base case:
        if (i2 - i1 == 1 && v[i2] < v[i1])
            Swap(v[i2], v[i1]);
    }
    else {         // Recursive case:
        p = Split(v, i1, i2);
        RecursiveSort(v, i1, p-1);
        RecursiveSort(v, p+1, i2);
    }
}

#else

template <class SOMETYPE>
void QUICKSORT<SOMETYPE>::RecursiveSort (apvector<SOMETYPE> &v,
                                                 int i1, int i2)

// Modified to assure the maximum recursion depth is O(log n).

{
    int p, m;

    while (i2 - i1 >= 1) {
        if (i2 - i1 == 1) { // Base case
            if (v[i2] < v[i1])
                Swap(v[i2], v[i1]);
            break;
        }
        else {  // Recursive case
            p = Split(v, i1, i2);
            m = (i2 + i1) / 2;
            if (p > m) {
                RecursiveSort(v, p+1, i2);
                                      // Process recursively the
                                      //   right part
                i2 = p-1;             // Continue iteratively with
                                      //   the left part
            }
            else {
                RecursiveSort(v, i1, p-1);
                                      // Process recursively the
                                      //   left part
                i1 = p+1;             // Continue iteratively with
                                      //   the right part
            }
        }
    }
}

#endif

//****************************************************************

template <class SOMETYPE>
int QUICKSORT<SOMETYPE>::Split(apvector<SOMETYPE> &v, int i1, int i2)

// Takes one of the values (here v[(i2+i1)/2]) as "pivot"
//   and splits the segment of the array v between i1 and i2
//   into two parts so that:
//      v[p] == pivot
//      v[j] <= pivot for j < p
//      v[j] >= pivot for j > p
// Returns p;

{
    int p = (i2 + i1) / 2;  // In this version the "pivot" value is the
                            //   middle element

    SOMETYPE pivot = v[p];

    int j1 = i1, j2 = i2;

    // Increment j1 if v[j1] <= pivot; decrement j2 if v[j2] >= pivot;
    //   otherwise swap v[j1] and v[j2] and advance both j1 and j2:
    while (j1 <= j2) {
        if (v[j1] <= pivot)
            j1++;
        else if (v[j2] >= pivot)
            j2--;
        else {
            Swap(v[j1], v[j2]);
            j1++;
            j2--;
        }
    }

    // Now v[j1] and v[j2] are adjacent elements; j1 > j2.
    //   Swap v[j1] or v[j2] with the pivot, making sure that
    //   the swapped element remains on the same side of the pair:
    if (p < j2) {
        Swap(v[p], v[j2]);
        p = j2;
    }
    else if (p > j1) {
        Swap(v[p], v[j1]);
        p = j1;
    }
    return p;
}

//****************************************************************
//***********                Heapsort               **************
//****************************************************************

// HEAPSORT.H

template <class SOMETYPE>
class HEAPSORT {

  public:

    void Sort(apvector<SOMETYPE> &v);
       // Sorts elements v[1] through v[n].  v[0] is not used.

  private:

    void Reheap(apvector<SOMETYPE> &v, int i, int n);

};

//****************************************************************

template <class SOMETYPE>
void HEAPSORT<SOMETYPE>::Reheap(apvector<SOMETYPE> &v, int i, int n)

//  Fixes the (sub)heap rooted at the node v[i], assuming that
//    all the nodes below it (all its descendants) already
//    satisfy the heap ordering property.
//  n is the largest possible index of an element in the heap:
//    v[j] is not in the heap when j > n (v[0] is unused).

{
    SOMETYPE x = v[i];  // Save the root value.

    int iParent = i;
    i *= 2;             // Set i to left child.
    while (i <= n) {
        // Set i to the right child, if it is larger:
        if (i < n && v[i] < v[i+1])
            i++;

        if (v[i] <= x)
            break;

        v[iParent] = v[i];     // Move the child up
        iParent = i;           // v[iParent] is now vacant
        i *= 2;                // i set to its left child
    }
    v[iParent] = x;
}

//****************************************************************

template <class SOMETYPE>
void HEAPSORT<SOMETYPE>::Sort(apvector<SOMETYPE> &v)

// Sorts elements v[1] through v[n]in ascending order
//   (v[0] is not used).

{
    int i, n = v.length() - 1;
    SOMETYPE x;

    // Fix heaps rooted in all nodes with children, going
    //   from the lowest-rightmost such node and proceeding
    //   left and up to the root.  v[i] is a leaf for any i > n/2,
    //   and v[n/2] is the parent of v[n], and therefore not a leaf.
    //   So we start with v[n/2].

    for (i = n/2;   i >= 1;   i--)
        Reheap(v, i, n);

    // Remove elements from the top of the heap and store them
    //   in reverse order starting from the end of the array:
    while (n > 1) {
        Swap(v[1], v[n]);
        n--;
        Reheap(v, 1, n);
    }
}

//****************************************************************
//***********                Test program           **************
//****************************************************************

#include <iostream.h>

static void Print (apvector<int> &v)

{
    int i;

    for (i = 0;   i < v.length();   i++)
        cout << v[i] << ' ';
    cout << endl;
}

//****************************************************************

int main()

{
    const int SIZE = 6;
    apvector<int> a(SIZE), v(SIZE);
    int target = 5, i;

    cout << "Enter " << SIZE << " integers ==> ";
    for (i = 0;   i < SIZE;   i++)
        cin >> a[i];
    Print (a);

    v = a;
    SelectionSort(v);
    cout << "Selection Sort:   ";
    Print(v);

    v = a;
    InsertionSort(v);
    cout << "Insertion Sort:   ";
    Print(v);

    v = a;
    InsertionSort2(v);
    cout << "Insertion Sort 2: ";
    Print(v);

    v = a;
    BubbleSort(v);
    cout << "Bubble Sort:      ";
    Print(v);

    v = a;
    MERGESORT<int> mergesort;
    mergesort.Sort(v);
    cout << "MERGESORT:        ";
    Print(v);

    v = a;
    QUICKSORT<int> quicksort;
    quicksort.Sort(v);
    cout << "QUICKSORT:        ";
    Print(v);

    v = a;
    // Shift to the right by one (v[0] is not used):
    v.resize(SIZE + 1);
    for (i = SIZE-1;   i >= 0;   i--)
        v[i+1] = v[i];

    HEAPSORT<int> heapsort;
    heapsort.Sort(v);

    // Shift to the left by one:
    for (i = 0;   i < SIZE;   i++)
        v[i] = v[i+1];
    v.resize(SIZE);

    cout << "HEAPSORT:         ";
    Print(v);

    cout << "Binary Search " << target << ": "
         << BinarySearch(v, target) << endl;

    return 0;
}

//****************************************************************

